/*
** read XML gpx file from stdin, extract and print to stdout
** trackpoint timestamp, latitude, longitude, altitude and range values
*/
#define _XOPEN_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <mxml.h>

static int trackpoints(double *);
static int trkpt(mxml_node_t *, mxml_node_t *, double *);
static double haversine(double, double, double, double);
static double bearing(double, double, double, double);

int main(int argc, char *argv[])
{
  int		status = 1;	/* 1 = OK, 0 = error(s) */
  double	origin[3], *porigin = NULL;

  switch (argc)
  {
    case 1:
      porigin = NULL;
      break;

    case 4:
      /* Latitude of Origin location, as GPS value */
      if (sscanf(argv[1],"%lf",origin+0) != 1)
      {
	fprintf(stderr,"Invalid origin latitude \"%s\"\n",argv[1]);
	status = 0;
      }

      /* Longitude of Origin location, as GPS value */
      if (sscanf(argv[2],"%lf",origin+1) != 1)
      {
	fprintf(stderr,"Invalid origin longitude \"%s\"\n",argv[2]);
	status = 0;
      }

      /* Range from Origin location, in metres */
      if (sscanf(argv[3],"%lf",origin+2) != 1)
      {
	fprintf(stderr,"Invalid range \"%s\"\n",argv[3]);
	status = 0;
      }
      if (status) porigin = origin;
      break;

    default:
      status = 0;	/* didnt get expected # of arguments */
      break;
  }

  if (status)
    status = trackpoints(porigin);
  else
    fprintf(stderr,"Usage: %s [lat lon range]\n",argv[0]);

  return status ? EXIT_SUCCESS : EXIT_FAILURE;
}

static int trackpoints(double *origin)
{
  int status = 0;
  mxml_node_t *xml;

  if (xml = mxmlLoadFile(NULL,stdin,MXML_TEXT_CALLBACK))
  {
    mxml_node_t *trk;

    /*
    ** ALL trk IN THIS xml
    */
    for (trk = mxmlFindElement(xml,xml,"trk",NULL,NULL,MXML_DESCEND);
	 trk;
	 trk = mxmlFindElement(trk,xml,"trk",NULL,NULL,MXML_NO_DESCEND))
    {
      mxml_node_t *trkseg;

      /*
      ** At this point, if we are inclined to, 
      ** we can pick out the <name> element associated with this <trk>
      ** and find a way to populate it in the output CSV
      */

      /*
      ** ALL trkseg IN THIS trk
      */
      for (trkseg = mxmlFindElement(trk,xml,"trkseg",NULL,NULL,MXML_DESCEND);
	   trkseg;
	   trkseg = mxmlFindElement(trkseg,xml,"trkseg",NULL,NULL,MXML_NO_DESCEND))
      {
	int trkwaypts;

	trkwaypts = trkpt(xml,trkseg,origin);

	status = trkwaypts || status;
      }
    }
    mxmlDelete(xml);

    if (status == 0)
      if (origin)
        fputs("No in-range GPX trackpoints found in XML file\n",stderr);
      else
        fputs("No GPX trackpoints found in XML file\n",stderr);
  }
  else fputs("No valid XML file on stdin\n",stderr);

  return status;
}

static int trkpt(mxml_node_t *xml, mxml_node_t *trkseg, double *origin)
{
  int foundtrackpoints = 0;
  mxml_node_t *trkpt = NULL;
  time_t p_tick = 0;
  double p_lat, p_lon, p_alt;

  /*
  ** ALL trkpt IN THIS trkseg
  */
  for (trkpt = mxmlFindElement(trkseg,xml,"trkpt",NULL,NULL,MXML_DESCEND);
       trkpt;
       trkpt = mxmlFindElement(trkpt,xml,"trkpt",NULL,NULL,MXML_NO_DESCEND))
  {
    int complete = 1;
    double t_lat, t_lon, t_alt, t_range,
	   d_bearing,
	   d_distance,
	   d_velocity;
    time_t t_tick,
	   d_tick;

    mxml_node_t *node;
    const char *attr;

    if (attr = mxmlElementGetAttr(trkpt,"lat"))
      sscanf(attr,"%lf",&t_lat);
    else
      complete = 0;

    if (attr = mxmlElementGetAttr(trkpt,"lon"))
      sscanf(attr,"%lf",&t_lon);
    else
      complete = 0;

    if (node = mxmlFindElement(trkpt,xml,"ele",NULL,NULL,MXML_DESCEND))
      sscanf(mxmlGetText(node,NULL),"%lf",&t_alt);
    else
      complete = 0;

    if (node = mxmlFindElement(trkpt,xml,"time",NULL,NULL,MXML_DESCEND))
    {
      struct tm TrackTime = {0,0,0,0,0,0,0,0,0};

      strptime(mxmlGetText(node,NULL),"%Y-%m-%dT%H:%M:%SZ",&TrackTime);
      t_tick = mktime(&TrackTime);
    }
    else complete = 0;

    /* derived information */
    if (origin)
    {
      t_range = haversine(origin[0], origin[1], t_lat, t_lon);
      if (t_range > origin[2]) complete = 0;
    }

    if (p_tick)
    {
      d_bearing = bearing(p_lat,p_lon,t_lat,t_lon);
      d_tick = t_tick - p_tick;
      {
	double dH, dV;

	dH = haversine(p_lat,p_lon,t_lat,t_lon);
	dV = t_alt - p_alt;
	d_distance = sqrt((dH*dH)+(dV*dV));
      }
      d_velocity = 3.6 * d_distance/d_tick;
    }

    if (complete)
    {
      /*
      ** Cell 1 - Unix timestamp of observation
      ** Cell 2 - decimal latitude of observation
      ** Cell 3 - decimal longitude of observation
      ** Cell 4 - decimal altitude (in metres) of observation
      */
      printf("%ld,%3.6lf,%3.6lf,%.2lf",(long int)t_tick,t_lat,t_lon,t_alt);

      /* Cell 5 - distance (in metres) from origin, or NULL */
      putchar(',');
      if (origin) printf("%.2lf",t_range);

      /* Cell 6 - direction of travel - compass bearing, or NULL */
      putchar(',');
      if (p_tick) printf("%.0lf",d_bearing);

      /* Cell 7 - Distance from previous point, or NULL*/
      putchar(',');
      if (p_tick) printf("%.0lf",d_distance);

      /* Cell 8 - Velocity in Km/Hr, or NULL */
      putchar(',');
      if (p_tick) printf("%.0lf",d_velocity);

      puts("");

      foundtrackpoints = 1;
      p_tick = t_tick;
      p_lat  = t_lat;
      p_lon  = t_lon;
      p_alt  = t_alt;
    }
  }

  return foundtrackpoints;
}

/*
** compute distance in metres between two (lat,lon) points
** using haversine formula
**
** origin_lat is point 1 latitude, origin_lon is point 1 longitude
** target_lat is point 2 latitude, target_lon is point 2 longitude
**
** R is the mean radius of the Earth in metres
** RADperDEG is the conversion ratio of degrees to radians
** NB:	to convert from degrees to radians, MULTIPLY degrees by RADperDEG
**	to convert from radians to degrees, DIVIDE radians by RADperDEG
**
** haversine computation derived from formula found at
** https://en.wikipedia.org/wiki/Haversine_formula
** See also:
**   https://www.movable-type.co.uk/scripts/gis-faq-5.1.html
**   http://mathforum.org/library/drmath/view/51879.html
**   https://rosettacode.org/wiki/Haversine_formula
*/

#define R 6371000
#define RADperDEG (M_PI / 180.)

static double haversine(double origin_lat, double origin_lon, double target_lat, double target_lon)
{
  double slon, slat;

  origin_lat *= RADperDEG; origin_lon *= RADperDEG;
  target_lat *= RADperDEG; target_lon *= RADperDEG;

  slon = sin((target_lon - origin_lon)/2.);
  slat = sin((target_lat - origin_lat)/2.);

  return 2. * R * asin(sqrt((slat * slat) + cos(origin_lat) * cos(target_lat)* slon * slon));
}

/*
** compute "forward azimuth" compass bearing from two pair of latitude,longitude points
** NB:	0 == North, 90 = East, 180 = South, 270 = West
**
** https://www.movable-type.co.uk/scripts/latlong-nomodule.html
** https://www.igismap.com/formula-to-find-bearing-or-heading-angle-between-two-points-latitude-longitude/
*/
static double bearing(double origin_lat, double origin_lon, double target_lat, double target_lon)
{
  double delta_lon,	/* delta longitude */
	 x, y;

  /* convert degrees to radians */
  origin_lat *= RADperDEG; origin_lon *= RADperDEG;
  target_lat *= RADperDEG; target_lon *= RADperDEG;

  delta_lon = target_lon - origin_lon;
  y = sin(delta_lon) * cos(target_lat);
  x = cos(origin_lat) * sin(target_lat) - sin(origin_lat) * cos(target_lat) * cos(delta_lon);

  /* convert radians to degrees, and degrees to compass heading */
  return fmod(360. + (atan2(y,x)/RADperDEG), 360.);	
}

/* $Id: trkpoints.c,v 1.14 2020/06/10 01:06:00 lpitcher Exp lpitcher $

    Copyright (C) 2020 Lew Pitcher <lew.pitcher@digitalfreehold.ca>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
